{
    This file is part of Chentrah,
    Copyright (C) 2004-2014 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}

{$mode objfpc}
{$longstrings off}
{$typedaddress on}

//{$define chbt_use_writeln}

unit chebs_heap_brutal_trace;

interface

const
  chbt_UseMegabytes = 500; //if the program runs out of this amout, it's over
    //because memory is never released, to trace illegal writes to the freed memory.

{ Brutal Tracer first allocates a solid block, then begins adding memory blocks
  from its end and info blocks from its beginning. Freed blocks are never reused.
  Any reallocation _always_ consists of freeing then allocating again, so adding
  array elements one by one will drain the pool rapidly. I recommend
  . }

  chbt_PaddingBytes = 32; //the padding between memory blocks that is filled
      // with $55 and monitored for changes

  chbt_AlignBytes = {$ifdef cpu64} 32 {$else} 16 {$endif} ;

  chbt_StackFrames = 15; //each info block contains a backtrace

  chbt_Reserve = 100000; //Bytes. To allow halting the program
     //without the tracer itself crashing



var
  chbt_SuspendChecks: boolean = false;

  chbt_EraseFreedBlocks: boolean = false; //convenient in that it guarantees
    //instant crashes at calls to destroyed class instances

  //Set this to 1 (or call chbt)StartChecking()) to pinpoint _where_ the memory is being trashed
  chbt_CheckEverythingWithEveryXOperations: longint = 1000;

procedure chbt_PerformCheck(); //To be able to trigger checks manually
                                  //It raises exception in case of trashed heap


procedure chbt_DumpUnfreedBlocksToFile(); //Use at the program exit to catch leaks

procedure chbt_StartChecking();

implementation

{$ifndef cge}
  uses lineinfo;
{$endif}

type
  PBacktraceRec = ^TBacktraceRec;
  TBacktraceRec = array[0..chbt_StackFrames - 1] of ptruint;

  PInfoBlock = ^TInfoBlock;
  TInfoBlock = record
    num: integer;
    mblock: pointer;
    size: ptruint;
    freed: boolean;
    freed_checksum: qword;
    previous: PInfoBlock;
    allocatedby, freedby: TBacktraceRec;
    thrashed: boolean;
    //checksum: qword; TODO: add info block checksum checking in case these are trashed
  end;

var
  pool, poolEnd, LastMb: pointer;
  Ib: PInfoBlock;
  Blocks: integer;
  chbt_file: Text;
  chbt_FileAssigned: boolean;
  checkCounter: integer;
  prevbt: TBacktraceRec;

procedure Crash(message: shortstring; info: PInfoBlock; var backtrace: TBacktraceRec
  ; pprev_backtrace: PBacktraceRec = nil
  ; show_previous: boolean = false; show_next: boolean = false); forward;

function IntToHex(Value: ptruint) : shortstring; overload;
var
  i: integer;
  b: ptruint;
begin
  SetLength(Result, sizeof(ptruint) * 2);
  for i:= 0 to sizeof(ptruint) * 2 - 1 do begin
    b:= (value shr (i * 4)) and $f;
    if b < $a
      then Result[sizeof(ptruint) * 2 - i]:= Chr(ord('0') + b)
      else Result[sizeof(ptruint) * 2 - i]:= Chr(ord('A') + b - $a);
  end;
end;

function IntToHex(Value: qword) : shortstring; overload;
var
  i: integer;
  b: ptruint;
begin
  SetLength(Result, sizeof(qword) * 2);
  for i:= 0 to sizeof(qword) * 2 - 1 do begin
    b:= (value shr (i * 4)) and $f;
    if b < $a
      then Result[sizeof(qword) * 2 - i]:= Chr(ord('0') + b)
      else Result[sizeof(qword) * 2 - i]:= Chr(ord('A') + b - $a);
  end;
end;



procedure GetBacktrace(var b: TBacktraceRec);
var
  bp, oldbp: pointer;
  i, skip: integer;
begin
  FillChar(b, sizeof(b), 0);
  bp:=get_caller_frame(get_frame);
  i:= 0;
  skip:= 0;
  while bp <> nil do begin
    if (i - skip) > High(b) then break;
    if i >= skip then b[i - skip]:= ptruint(get_caller_addr(bp));
    inc(i);
    oldbp:= bp;
    bp:=get_caller_frame(bp);
    if (bp<=oldbp) or (bp>(StackBottom + StackLength)) then
      bp:=nil;
  end;
end;

function Align(p: ptruint): ptruint;
begin
  Result:= (p div chbt_AlignBytes) * chbt_AlignBytes;
end;



function IntToStr(value: integer): shortstring;
var
  i, v, m: integer;
  a: array[0..10] of byte;
  b: boolean;
begin
  if value = 0 then Exit('0');
  Result:= '';
  if (value < 0) then begin
    Result:= '-';
    value:= -value;
  end;
  v:= value;
  for i:= 0 to 10 do begin
    a[i]:= v mod 10;
    v:= v div 10;
  end;
  b:= false;
  for i:= 10 downto 0 do begin
    if (a[i] = 0) and not b then continue;
    b:= true;
    Result+= chr(ord('0') + a[i]);
  end;
end;

function DescribeAddr(p: ptruint): shortstring;
var
   func,
   source: shortstring;
   line: longint;
begin
{$ifdef cge}
  Result:= BackTraceStrFunc(pointer(p));
//  Result:= IntToHex(p) + ' ' + ExplainLineInfo(pointer(p));
{$else}
if GetLineInfo(p, func, source, line) and (line <> 0) and (source <> '')
  then Result:= func + ' (' + IntToStr(line) + ' of ' + source + ')'
  else Result:= IntToHex(p);
{$endif}
end;


function Alloc(oname: shortstring; size: ptruint; var bt: TBacktraceRec): integer;
//it either works or terminates the entire program.
begin
  Inc(Blocks);
  Result:= Blocks;
  FillChar(Ib[Blocks], SizeOf(Ib[Blocks]), 0);
  Ib[Blocks].mblock:= pointer(Align(ptruint(LastMb) - size - chbt_PaddingBytes));
  Ib[Blocks].num:= Blocks;
  Ib[Blocks].Size:= size;
  Ib[Blocks].AllocatedBy:= bt;
  if not chbt_SuspendChecks then begin
    if Ib[Blocks].mblock < @(Ib[blocks + 1]) + 2 * chbt_PaddingBytes + chbt_AlignBytes + chbt_Reserve
      then Crash(oname + ': Ran out of memory pool', nil, bt);
  end;
  if Ib[Blocks].mblock <= @(Ib[blocks + 1]) + 2 * chbt_PaddingBytes + chbt_AlignBytes then begin
    if chbt_FileAssigned then begin
      {$ifdef chbt_use_writeln}
      WriteLn('RAN OUT OF MEMORY POOL');
      {$endif}
      WriteLn(chbt_file, 'RAN OUT OF MEMORY POOL');
    end;
    RunError(203); //halt with 'Heap overflow error'
  end;
  if (blocks > 1) then begin
    //fill the padding space between this block and the next one
    FillChar((Ib[Blocks].mblock + Ib[Blocks].Size)^, ptruint(Ib[Blocks - 1].mblock) - ptruint(Ib[Blocks].mblock) - Ib[Blocks].Size, $55);
  end;

  //fill the padding space before this block
  FillChar((Ib[Blocks].mblock - chbt_PaddingBytes)^, chbt_PaddingBytes, $55);
  LastMb:= Ib[Blocks].mblock;
  {$ifdef chbt_use_writeln}
  writeln('Allocated #', Blocks,' at ',IntToHex(ptruint(Ib[Blocks].mblock)));
  {$endif}
end;


function GetChecksum(p: pointer; size: integer): qword;
var i: integer;
begin
  Result:= 0;
  for i:= 0 to size - 1 do
    Result:= Result xor (qword(byte((p + i) ^)) shl ((i and 3) * 8));
end;

procedure Freeze(ibnum: integer; var bt: TBacktraceRec);
begin
  Ib[ibnum].freed:= true;
  if chbt_EraseFreedBlocks then FillChar(Ib[ibnum].mblock^, Ib[ibnum].size, $55);
  Ib[ibnum].freed_checksum:= GetChecksum(Ib[ibnum].mblock, Ib[ibnum].size);
  Ib[ibnum].freedby:= bt;
  {$ifdef chbt_use_writeln}
  writeln('Deallocated #', ibnum,' , ',IntToHex(ptruint(Ib[ibnum].mblock)));
  {$endif}
end;

procedure CheckOneBlock(ibnum: integer; var bt: TBacktraceRec);
var
  i, size: integer;
  p: pbyte;
begin
  if Ib[ibnum].freed then begin
    if Ib[ibnum].freed_checksum <> GetChecksum(Ib[ibnum].mblock, Ib[ibnum].size)
      then Crash('A change in a DEALLOCATED memory block detected '
      //+ IntToHex( Ib[ibnum].freed_checksum) + ' ' + IntToHex(GetChecksum(Ib[ibnum].mblock, Ib[ibnum].size))
      , @Ib[ibnum], bt, @prevbt);
  end;

  if ibnum = Blocks then begin
    //the last block should be checked for underflow
    p:= pbyte(Ib[ibnum].mblock - chbt_PaddingBytes);
    for i:= 1 to chbt_PaddingBytes - 1 do
      if p[i] <> $55
        then Crash('Memory block UNDERFLOW detected', @Ib[ibnum], bt, @prevbt, false, false);
  end;

  if ibnum < 2 then exit;
  //TODO: add underflow checking for the last block!
  size:= ptruint(Ib[ibnum - 1].mblock - (Ib[ibnum].mblock + Ib[ibnum].Size));
  p:= pbyte(Ib[ibnum].mblock + Ib[ibnum].Size);
  for i:=0 to size - 1 do
    if p[i] <> $55 then
      if i < (size div 2)
        then Crash('Memory block OVERFLOW detected', @Ib[ibnum], bt, @prevbt, false, true)
        else Crash('Memory block UNDERFLOW detected', @Ib[ibnum - 1], bt, @prevbt, true, false);
end;

procedure PerformCheck(var bt: TBacktraceRec);
var i: integer;
begin
  if chbt_SuspendChecks then Exit;
  inc(checkCounter);
  if checkCounter < chbt_CheckEverythingWithEveryXOperations then Exit;
  checkCounter:= 0;
//writeln('checking ', Blocks, ' blocks...');
  for i:= 1 to Blocks do CheckOneBlock(i, bt);
  prevbt:= bt;
end;

procedure chbt_PerformCheck(); //To be able to trigger checks manually
var                                 //It raises exception in case of trashed heap
  bt: TBacktraceRec;
begin
  GetBacktrace(bt);
  checkCounter:= chbt_CheckEverythingWithEveryXOperations;
  PerformCheck(bt);
end;

procedure chbt_StartChecking();
begin
  chbt_CheckEverythingWithEveryXOperations:= 1;
end;

//TODO: optimize this shit!
function FindIb(oname: shortstring; p: pointer; var bt: TBacktraceRec): integer;
var
  i: integer;
begin
  if (p >= poolEnd) or (p < LastMb)
    then Crash(oname + ': the pointer provided is outside the allocated pool!', nil, bt);
  For i:= Blocks downto 0 do
    if Ib[i].mblock = p then Exit(i);
  Crash(oname + ': the pointer ' + IntToHex(ptruint(p)) + ' is not a memory block', nil, bt);
end;


procedure LogBacktrace(var bt: TBacktraceRec);
var i: integer;
begin
  for i:= 0 to high(bt) do begin
    if bt[i] = 0 then Exit;
    {$ifdef chbt_use_writeln}
    WriteLn('    ' + DescribeAddr(bt[i]));
    {$endif}
    WriteLn(chbt_file, '    ' + DescribeAddr(bt[i]));
  end;
end;

procedure LogInfoBlock(info: PInfoBlock; logNeighbors: boolean = true);
begin
  {$ifdef chbt_use_writeln}
  WriteLn('Memory block #', info^.num, ':');
  WriteLn('  Size: ', info^.size);
  {$endif}
  WriteLn(chbt_file, 'Memory block #', info^.num, ':');
  WriteLn(chbt_file, '  Size: ', info^.size);
  if info^.allocatedby[0] > 0 then begin
    {$ifdef chbt_use_writeln}
    WriteLn('  Allocated by:');
    {$endif}
    WriteLn(chbt_file, '  Allocated by:');
    LogBacktrace(info^.allocatedby);
  end;
  if info^.freed then begin
    {$ifdef chbt_use_writeln}
    WriteLn('  Freed by:');
    {$endif}
    WriteLn(chbt_file, '  Freed by:');
    LogBacktrace(info^.freedby);
  end
end;

procedure Crash(message: shortstring; info: PInfoBlock; var backtrace: TBacktraceRec
  ; pprev_backtrace: PBacktraceRec = nil
  ; show_previous: boolean = false; show_next: boolean = false);
begin
  chbt_SuspendChecks:= true;
  if (chbt_FileAssigned) then begin
    {$ifdef chbt_use_writeln}
    WriteLn(message);
    {$endif}
    WriteLn(chbt_file, message);
    if Assigned(pprev_backtrace) then begin
      {$ifdef chbt_use_writeln}
      WriteLn('  Before:');
      {$endif}
      WriteLn(chbt_file, '  Before:');
      LogBacktrace(backtrace);
      {$ifdef chbt_use_writeln}
      WriteLn('  After:');
      {$endif}
      WriteLn(chbt_file, '  After:');
      LogBacktrace(pprev_backtrace^);
    end else begin
      {$ifdef chbt_use_writeln}
      WriteLn('  Backtrace:');
      {$endif}
      WriteLn(chbt_file, '  Backtrace:');
      LogBacktrace(backtrace);
    end;
    if Assigned(info) then begin
      LogInfoBlock(info);
      if (info^.num > 0) and show_previous then LogInfoBlock(@Ib[info^.num - 1]);
      if (info^.num < Blocks) and show_next then LogInfoBlock(@Ib[info^.num + 1]);
    end;
    LogBacktrace(backtrace);
  end;
  RunError(204); //halt with 'Invalid pointer operation'
end;

procedure chbt_DumpUnfreedBlocksToFile(); //Use at the program exit to catch leaks
var
  i, count: integer;
  b: boolean;
begin
  chbt_SuspendChecks:= true;
  count:= 0;
  for i:=1 to Blocks do begin
    if Ib[i].freed then continue;
    if count = 0 then begin
      {$ifdef chbt_use_writeln}
      WriteLn('Listing memory leaks:');
      {$endif}
      WriteLn(chbt_file, 'Listing memory leaks:');
    end;
    LogInfoBlock(@Ib[i], false);
    inc(count)
  end;
  {$ifdef chbt_use_writeln}
  if count = 0 then WriteLn('No memory leaks found')
               else WriteLn('Memory leaks: ',count,' blocks found');
  {$endif}
  if count = 0 then WriteLn(chbt_file, 'No memory leaks found')
               else WriteLn(chbt_file, 'Memory leaks: ',count,' blocks found');
end;

procedure Dump();
begin
  chbt_SuspendChecks:= true;

end;

Function chbt_GetMem  (Size : ptruint) : Pointer;
var
  bt: TBacktraceRec;
  ibi: integer;
begin
  {$ifdef chbt_use_writeln}
  writeln('GetMem(' + IntToStr(Size) + ')');
  {$endif}
  GetBacktrace(bt);
  PerformCheck(bt);
  ibi:= Alloc('GetMem', size, bt);
  Result:= Ib[ibi].mblock;
  FillChar(Result^, Size, $55);
end;

Function chbt_FreeMem (P : pointer) : ptruint;
var
  bt: TBacktraceRec;
  ibi: integer;
  info: PInfoBlock;
begin
  {$ifdef chbt_use_writeln}
  writeln('FreeMem(' + IntToHex(ptruint(p)) + ')');
  {$endif}

  GetBacktrace(bt);
  PerformCheck(bt);
  if p = nil then begin
    if chbt_SuspendChecks then Exit(0);
    Crash('FreeMem: attempt to free nil pointer', nil, bt);
  end;
  ibi:= FindIb('FreeMem', p, bt);
  if (ibi < 0) and chbt_SuspendChecks then Exit(0);
  if (Ib[ibi].freed) then begin
    if chbt_SuspendChecks then Exit(0);
    Crash('FreeMem: attempt to free an already freed block', @Ib[ibi], bt);
  end;
  Freeze(ibi, bt);
  Result:= 0;
end;

Function chbt_FreeMemSize(p:pointer;Size:ptruint):ptruint;
var
  bt: TBacktraceRec;
  ibi: integer;
  info: PInfoBlock;
begin
  {$ifdef chbt_use_writeln}
  writeln('FreeMemSize(' + IntToHex(ptruint(p)) + ', ' +IntToStr(Size) + ')');
  {$endif}

  GetBacktrace(bt);
  PerformCheck(bt);
  if (Size < 0) then begin
    if chbt_SuspendChecks then Exit(0);
    Crash('FreeMemSize: the size provided is less than 1', nil, bt);
  end;
  if p = nil then begin
    if chbt_SuspendChecks then Exit(0);
    Crash('FreeMemSize: attempt to free a nil pointer', nil, bt);
  end;
  ibi:= FindIb('FreeMemSize', p, bt);
  if (ibi < 0) and chbt_SuspendChecks then Exit(0);
  if (Ib[ibi].size <> Size) then begin
    if chbt_SuspendChecks then Exit(0);
    Crash('FreeMemSize: the size provided does not match block size', @Ib[ibi], bt);
  end;
  if (Ib[ibi].freed) then begin
    if chbt_SuspendChecks then Exit(0);
    Crash('FreeMemSize: attempt to free an already freed block', @Ib[ibi], bt);
  end;
  Freeze(ibi, bt);
  Result:= 0;
end;


Function chbt_AllocMem  (Size : ptruint) : Pointer;
var
  bt: TBacktraceRec;
  ibi: integer;
begin
//  writeln('AllocMem(' +IntToStr(Size) + ')');
  GetBacktrace(bt);
  PerformCheck(bt);
  ibi:= Alloc('AllocMem', size, bt);
  Result:= Ib[ibi].mblock;
  FillChar(Result^, Size, 0);
end;

Function chbt_ReAllocMem (var p:pointer; Size:ptruint):Pointer;
var
  bt: TBacktraceRec;
  ibi: integer;
  info: PInfoBlock;
  oldibi: integer = -1;
  oldinfo: ptruint;
begin
  {$ifdef chbt_use_writeln}
  writeln('ReallocMem(' + IntToHex(ptruint(p)) + ', ' +IntToStr(Size) + ')');
  {$endif}
  GetBacktrace(bt);
  PerformCheck(bt);
  if p <> nil then begin
    //ALWAYS allocate a new block freezing the old one.
    oldibi:= FindIb('ReAllocMem', p, bt);
    if (oldibi < 0) and chbt_SuspendChecks then
      p:= nil
    else
      if (Ib[oldibi].freed) then begin
        if chbt_SuspendChecks then begin
          p:= nil;
          Exit(nil);
        end;
        Crash('ReAllocMem: attempt to realloc an expired pointer', @Ib[oldibi], bt);
      end;
  end;

  if size = 0 then  begin
    if p = nil then begin
      if chbt_SuspendChecks then Exit(nil);
      Crash('ReAllocMem: attempt to realloc 0 bytes for a nil pointer', nil, bt);
    end;

    //the same as FreeMem;
    Freeze(oldibi, bt);
    p:= nil;
    Result:= nil;
  end else begin
    ibi:= Alloc('ReAllocMem', size, bt);
    Result:= Ib[ibi].mblock;
    p:= Result;
    if oldibi < 0 then FillChar(Result^, Size, $55);
  end;

  if (ibi >= 0) and (oldibi >=0) then begin
    //copy the memory content
    if Size < Ib[oldibi].size then
      Move(Ib[oldibi].mblock^, Result^, Size)
    else begin
      Move(Ib[oldibi].mblock^, Result^, Ib[oldibi].size);
      FillChar((Result + Ib[oldibi].size)^, Size - Ib[oldibi].size, $55);
    end;
    Freeze(oldibi, bt);
  end;
end;

Function chbt_MemSize (p:pointer): ptruint;
var
  bt: TBacktraceRec;
  ibi: integer;
begin
  {$ifdef chbt_use_writeln}
  writeln('MemSize(' + IntToHex(ptruint(p)) + ')');
  {$endif}
  GetBacktrace(bt);
  PerformCheck(bt);
  if p = nil then begin
    if chbt_SuspendChecks then Exit(0);
    Crash('MemSize: attempt to get the size of a nil pointer', nil, bt);
  end;
  ibi:= FindIb('MemSize', p, bt);
  if (ibi < 0) and chbt_SuspendChecks then Exit(0);
  if (Ib[ibi].freed) then begin
    if chbt_SuspendChecks then Exit(0);
    Crash('MemSize: attempt to get the size of an expired pointer', @Ib[ibi], bt);
  end;
  Result:= Ib[ibi].size;
end;

function chbt_GetHeapStatus:THeapStatus;
begin
  fillchar(Result ,sizeof(Result), 0);
end;

function chbt_GetFPCHeapStatus:TFPCHeapStatus;
begin
  fillchar(Result ,sizeof(Result), 0);
end;

Const
 chbt_MemoryManager : TMemoryManager =
    (
      NeedLock : true;
      GetMem : @chbt_Getmem;
      FreeMem : @chbt_FreeMem;
      FreememSize : @chbt_FreememSize;
      AllocMem : @chbt_AllocMem;
      ReallocMem : @chbt_ReAllocMem;
      MemSize : @chbt_MemSize;
      InitThread : nil;
      DoneThread : nil;
      RelocateHeap : nil;
      GetHeapStatus : @chbt_GetHeapStatus;
      GetFPCHeapStatus: @chbt_GetFPCHeapStatus;
    );

Var
  OldMemoryManager : TMemoryManager;

Initialization
  pool:= GetMem(ptruint(chbt_UseMegabytes) * 1024 * 1024);
  Ib:= pool;
  LastMb:= pool + ptruint(chbt_UseMegabytes) * 1024 * 1024;
  poolEnd:= LastMb;
  GetMemoryManager (OldMemoryManager);
  SetMemoryManager (chbt_MemoryManager);
  AssignFile(chbt_file, ParamStr(0) + '_chbt.txt');
  Rewrite(chbt_file);
  {$ifdef chbt_use_writeln}
  WriteLn('Cheb''s Brutal Heap Tracer started.');
  {$endif}
  WriteLn(chbt_file, 'Cheb''s Brutal Heap Tracer started.');
  chbt_FileAssigned:= true;


Finalization
//writeln('chbt finalization begin');
  if not chbt_SuspendChecks then begin
    {$ifdef chbt_use_writeln}
    WriteLn('Normal termination.');
    {$endif}
    WriteLn(chbt_file, 'Normal termination.');
  end;
  chbt_PerformCheck();
  if not chbt_SuspendChecks then chbt_DumpUnfreedBlocksToFile();
  SetMemoryManager (OldMemoryManager);
  FreeMem(pool);
  if chbt_FileAssigned then CloseFile(chbt_file);
//writeln('chbt finalization end');
end.
